#!/usr/bin/env python
# -*- coding: utf-8 -*-

__author__ = 'Ben Lopatin'
__email__ = 'ben@wellfire.co'
__version__ = '0.1.4'


from client import GeocodioClient
from data import Address, Location, LocationCollection

__all__ = [GeocodioClient, Address, Location, LocationCollection]



#!/usr/bin/env python
# -*- coding: utf-8 -*-


class Address(dict):
    """
    Dictionary class that provides some convenience wrappers for accessing
    commonly used data elements on an Address.
    """

    def __init__(self, address_dict):
        super(Address, self).__init__(address_dict)

    @property
    def accuracy(self):
        """
        Returns a tuple represneting the location of the address in a
        GIS coords format, i.e. (longitude, latitude).
        """
        try:
            return self["accuracy"]
        except KeyError:
            return None

    @property
    def coords(self):
        """
        Returns a tuple represneting the location of the address in a
        GIS coords format, i.e. (longitude, latitude).
        """
        try:
            return (self["location"]["lng"], self["location"]["lat"])
        except KeyError:
            return None


class Location(dict):
    """
    Dictionary class that provides some convenience accessors to commonly used
    data elements.
    """

    def __init__(self, result_dict):
        super(Location, self).__init__(result_dict)
        try:
            self.best_match = Address(self["results"][0])
        # A KeyError would be raised if an address could not be parsed or
        # geocoded, i.e. from a batch address geocoding process. An index error
        # would be raised under similar circumstances, e.g. the 'results' key
        # just refers to an empty list.
        except (KeyError, IndexError):
            self.best_match = Address({})

    @property
    def coords(self):
        """
        Returns a tuple represneting the location of the first result in a
        GIS coords format, i.e. (longitude, latitude).
        """
        return self.best_match.coords

    @property
    def accuracy(self):
        """
        Returns a tuple represneting the location of the first result in an
        everyday (latitude, longitude) format.
        """
        return self.best_match.accuracy


class LocationCollection(list):
    """
    A list of Location objects, with dictionary lookup by address.
    """
    lookups = {}

    def __init__(self, results_dict):
        """
        Loads the individual responses into an internal list and uses the query
        values as lookup keys.
        """
        results = []
        for index, result in enumerate(results_dict['results']):
            results.append(Location(result['response']))
            self.lookups[result['query']] = index
        super(LocationCollection, self).__init__(results)

    def get(self, key):
        """
        Returns an individual Location by formatted address lookup
        """
        return self[self.lookups[key]]

    @property
    def coords(self):
        """
        Returns a list of tuples for the best matched coordinates.
        """
        return [l.coords for l in self]



#!/usr/bin/env python
# -*- coding: utf-8 -*-


import json
import requests
from data import Address, Location, LocationCollection
from exceptions import (GeocodioAuthError, GeocodioDataError,
        GeocodioServerError, GeocodioError)


def error_response(response):
    """
    Raises errors matching the response code
    """
    if response.status_code >= 500:
        raise GeocodioServerError
    elif response.status_code == 403:
        raise GeocodioAuthError
    elif response.status_code == 422:
        raise GeocodioDataError(response.json()['error'])
    else:
        raise GeocodioError("Unknown service error (HTTP {0})".format(response.status_code))


class GeocodioClient(object):
    """
    Client connection for Geocod.io API
    """
    BASE_URL = "http://api.geocod.io/v1/{verb}"

    def __init__(self, key):
        self.API_KEY = key

    def parse(self, address):
        """
        Returns an Address dictionary with the components of the queried
        address.

        >>> client = GeocodioClient('some_api_key')
        >>> client.parse("1600 Pennsylvania Ave, Washington DC")
        {
            "address_components": {
                "number": "1600",
                "street": "Pennsylvania",
                "suffix": "Ave",
                "city": "Washington",
                "state": "DC"
            },
            "formatted_address": "1600 Pennsylvania Ave, Washington DC"
        }
        """
        url = self.BASE_URL.format(verb="parse")
        response = requests.get(url, params={'q': address, 'api_key': self.API_KEY})
        if response.status_code != 200:
            return error_response(response)
        return Address(response.json())

    def batch_geocode(self, addresses):
        """
        Returns an Address dictionary with the components of the queried
        address.
        """
        url = self.BASE_URL.format(verb="geocode")

        response = requests.post(url, params={'api_key': self.API_KEY},
                headers={'content-type': 'application/json'},
                data=json.dumps(addresses))
        if response.status_code != 200:
            return error_response(response)
        return LocationCollection(response.json())

    def geocode_address(self, address):
        """
        Returns a Location dictionary with the components of the queried
        address and the geocoded location.

        >>> client = GeocodioClient('some_api_key')
        >>> client.geocode("1600 Pennsylvania Ave, Washington DC")
        {
        "input": {
            "address_components": {
                "number": "1600",
                "street": "Pennsylvania",
                "suffix": "Ave",
                "city": "Washington",
                "state": "DC"
            },
            "formatted_address": "1600 Pennsylvania Ave, Washington DC"
        },
        "results": [
            {
            "address_components": {
                "number": "1600",
                "street": "Pennsylvania",
                "suffix": "Ave",
                "city": "Washington",
                "state": "DC",
                "zip": "20500"
            },
            "formatted_address": "1600 Pennsylvania Ave, Washington DC, 20500",
            "location": {
                "lat": 38.897700000000,
                "lng": -77.03650000000,
            },
            "accuracy": 1
            },
            {
            "address_components": {
                "number": "1600",
                "street": "Pennsylvania",
                "suffix": "Ave",
                "city": "Washington",
                "state": "DC",
                "zip": "20500"
            },
            "formatted_address": "1600 Pennsylvania Ave, Washington DC, 20500",
            "location": {
                "lat": 38.897700000000,
                "lng": -77.03650000000,
            },
            "accuracy": 0.8
            }
        ]
        }
        """
        url = self.BASE_URL.format(verb="geocode")
        response = requests.get(url, params={'q': address, 'api_key': self.API_KEY})
        if response.status_code != 200:
            return error_response(response)
        return Location(response.json())

    def geocode(self, address_data):
        """
        Returns geocoding data for either a list of addresses or a single
        address represented as a string.

        Provides a single point of access for end users.
        """
        if isinstance(address_data, list):
            return self.batch_geocode(address_data)
        return self.geocode_address(address_data)
